{ File   : CLUExt.dpr
  Author : Deniz Oezmen
  Created: 2005-12-02
  Changed: 2005-12-02

  Broken Sword 1 CLU extractor.
}
program CLUExt;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils;

type
  TFileInfo = packed record
                ID,
                Offset,
                Size  : Longword
              end;

  TSection  = packed record
                BaseID,
                NumFiles: Longword;
                Files   : array of TFileInfo
              end;

  TCLUInfo = packed record
               Name    : array[0..31] of Char;
               BaseID,
               NumSects: Longword;
               Sections: array of TSection
             end;

  TRIFInfo = packed record
               NumClusters: Longword;
               CLUInfo    : array of TCLUInfo
             end;

var
  RIFInfo: TRIFInfo;
  RIFFile,
  CLUFile: file;

function OpenRIFFile(var RIFFile: file; RIFName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(RIFFile, RIFName);

  {$I-}
  Reset(RIFFile, 1);
  {$I+}
  OpenRIFFile := (IOResult = 0)
end;

function ReadRIFInfo(var RIFInfo: TRIFInfo): Boolean;
var
  i, j, k  : Longint;
  BytesRead: Longword;
begin
  Result := False;

  with RIFInfo do
  begin
    // read number of clusters
    BlockRead(RIFFile, NumClusters, 4, BytesRead);
    if BytesRead < 4 then
      Exit;

    // read base cluster info
    SetLength(CLUInfo, NumClusters);
    for i := 0 to NumClusters - 1 do
    begin
      BlockRead(RIFFile, CLUInfo[i].BaseID, 4, BytesRead);
      if BytesRead < 4 then
        Exit
    end;

    // read info for each cluster
    for i := 0 to NumClusters - 1 do
      with CLUInfo[i] do
      begin
        // read name
        BlockRead(RIFFile, Name[0], 32, BytesRead);
        if BytesRead < 32 then
          Exit;

        // read number of sections
        BlockRead(RIFFile, NumSects, 4, BytesRead);
        if BytesRead < 4 then
          Exit;

        // read base section info
        SetLength(Sections, NumSects);
        for j := 0 to NumSects - 1 do
        begin
          BlockRead(RIFFile, Sections[j].BaseID, 4, BytesRead);
          if BytesRead < 4 then
            Exit
        end;

        // read info for each section
        for j := 0 to NumSects - 1 do
          with Sections[j] do
          begin
            // read number of files
            BlockRead(RIFFile, NumFiles, 4);

            // read file IDs
            SetLength(Files, NumFiles);
            for k := 0 to NumFiles - 1 do
            begin
              BlockRead(RIFFile, Files[k].ID, 4, BytesRead);
              if BytesRead < 4 then
                Exit
            end;

            // read file offsets and lengths
            for k := 0 to NumFiles - 1 do
            begin
              BlockRead(RIFFile, Files[k].Offset, 4, BytesRead);
              if BytesRead < 4 then
                Exit;
              BlockRead(RIFFile, Files[k].Size, 4, BytesRead);
              if BytesRead < 4 then
                Exit
            end
          end
      end
  end;

  Result := True
end;

function OpenCLUFile(var CLUFile: file; CLUName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(CLUFile, CLUName);

  {$I-}
  Reset(CLUFile, 1);
  {$I+}
  OpenCLUFile := (IOResult = 0)
end;

function Prefix(FileName: string): string;
var
  i: Longint;
begin
  FileName := ExpandFileName(FileName);

  repeat
    i := Pos('\', FileName);
    FileName := Copy(FileName, i + 1, Length(FileName) - i)
  until i = 0;

  i := Length(FileName) + 1;
  repeat
    Dec(i)
  until (i = 0) or (FileName[i] = '.');

  if i > 0 then
    Result := Copy(FileName, 1, i - 1)
  else
    Result := FileName
end;

function ValidateCLU(CLUName: string): Longint;
var
  i: Longint;
begin
  Result := -1;

  CLUName := AnsiUpperCase(Prefix(CLUName));

  // try to find cluster file in list
  for i := 0 to RIFInfo.NumClusters - 1 do
    if AnsiUpperCase(RIFInfo.CLUInfo[i].Name) = CLUName then
    begin
      Result := i;
      Break
    end
end;

procedure ExtractFiles(CLUNum: Longword);
var
  j,
  SectionNum,
  FileNum   : Longint;
  FileName  : string;
  OutFile   : file;
  Buffer    : array[0..65535] of Byte;
begin
  with RIFInfo.CLUInfo[CLUNum] do
    for SectionNum := 0 to NumSects - 1 do
      with Sections[SectionNum] do
        for FileNum := 0 to NumFiles - 1 do
          with Files[FileNum] do
          begin
            FileName := IntToHex(ID, 8) + '.' + IntToStr(SectionNum);

            Writeln('Extracting ', FileName, ' @ ', Offset, ' (', Size,
              ' bytes)');

            Seek(CLUFile, Offset);
                                                     
            FileMode := fmOpenReadWrite;
            AssignFile(OutFile, FileName);
            Rewrite(OutFile, 1);

            for j := 1 to Size div SizeOf(Buffer) do
            begin
              BlockRead(CLUFile, Buffer, SizeOf(Buffer));
              BlockWrite(OutFile, Buffer, SizeOf(Buffer))
            end;
            BlockRead(CLUFile, Buffer, Size mod SizeOf(Buffer));
            BlockWrite(OutFile, Buffer, Size mod SizeOf(Buffer));

            CloseFile(OutFile)
          end
end;

procedure ParseFiles(RIFName, FilePara: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  CLUName  : string;
  SearchRec: TSearchRec;
  PFilePara: PChar;
begin
  // first check for the RIF file
  if OpenRIFFile(RIFFile, RIFName) then
  begin
    if not ReadRIFInfo(RIFInfo) then
    begin
      WriteLn(RIFName, ' is not a valid RIF file');
      Exit
    end
  end
  else
  begin
    WriteLn('Could not open', RIFName);
    Exit
  end;

  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  FileMode := fmOpenRead;

  // search for all files matching FilePara
  if FindFirst(PFilePara, faAnyFile - faDirectory - faVolumeID,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      CLUName := Dir + FileName;

      // if this file is accessible, analyze it
      if OpenCLUFile(CLUFile, CLUName) then
      begin
        if ValidateCLU(CLUName) >= 0 then
        begin
          Writeln('Extracting from ', CLUName, ' ...');
          ExtractFiles(ValidateCLU(CLUName));
          Close(CLUFile)
        end
        else
          WriteLn(CLUName, ' does not match ', RIFName)
      end
      else
        Writeln(CLUName, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, ' not found')
end;

begin
  if ParamCount = 2 then
    ParseFiles(ParamStr(1), ParamStr(2))
  else
    Writeln('Incorrect number of parameters.')
end.

